"""
/***************************************************************************
 LandXML2QGIS
                                 A QGIS plugin
 Import Vic, NSW LandXML into QGIS temp layers
 Generated by Plugin Builder: http://g-sherman.github.io/Qgis-Plugin-Builder/
                              -------------------
        begin                : 2019-08-22
        git sha              : $Format:%H$
        copyright            : (C) 2019 by James Leversha, Department of Environment, Land, Water and Planning
        email                : james.k.leversha@delwp.vic.gov.au
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""

import math
from ..dcmvr import ts_functions
import shapely.geometry as sg


def calc_arc_values(start_neh, end_neh, radius, rot_dir, arc_length):
    try:

        radius = float(radius)
        arc_length = float(arc_length)

        start_neh_sg = start_neh
        end_neh_sg = end_neh
        start_neh = start_neh.coords[:][0]
        end_neh = end_neh.coords[:][0]
        start_e, start_n, start_h = None, None, None
        end_e, end_n, end_h = None, None, None
        height_dif = None
        if len(start_neh) == 2 or len(end_neh) == 2:
            start_n = start_neh[1]
            end_n = end_neh[1]
            start_e = start_neh[0]
            end_e = end_neh[0]
        elif len(start_neh) == 3 and len(end_neh) == 3:
            start_n = start_neh[1]
            end_n = end_neh[1]
            start_e = start_neh[0]
            end_e = end_neh[0]
            start_h = start_neh[2]
            end_h = end_neh[2]
            height_dif = start_h - end_h

        if start_e is not None or end_e is not None:
            lines = []
            # calculate if arc is greater than 180
            alt_cent = False
            if radius is not None:
                radius = float(radius)
                half_circum = math.pi * radius
                if arc_length > half_circum:
                    alt_cent = True
                    rot_dir = flip_rotation(rot_dir)
            # calculate the scale factor for the radius
            chord = arc2chord(arc_length, radius)
            distance = calc_distance(start_neh_sg, end_neh_sg)
            csf = distance / chord

            bearing = calc_bearing(end=end_neh_sg, start=start_neh_sg)

            bearing = math.degrees(bearing)

            nradius = float(radius) * 1  # csf
            if nradius < (distance / 2.):
                nradius = radius

            bearing = calc_inside_360(bearing)
            rot_val = calc_value_for_rotation(rot_dir)

            angle_bw_points, bearing_to_centre, nradius = calc_centre_point_of_curve(nradius, bearing,
                                                                                     distance, rot_val)

            delta_east, delta_north = calc_new_point(bearing_to_centre, nradius)
            centre_point = start_e + delta_east, start_n + delta_north

            if bearing_to_centre < 180:
                centre_to_start = bearing_to_centre + 180
            else:
                centre_to_start = bearing_to_centre - 180

            if alt_cent is True:
                angle_bw_points = 360 - angle_bw_points
                rot_dir = flip_rotation(rot_dir)
            rot_val = calc_value_for_rotation(rot_dir)

            lines.append([start_e, start_n])
            sum_of_angles = 0
            d = .001  # max deviation from arc
            a = nradius * nradius
            b = (nradius - d) * (nradius - d)
            dist_of_change = 2 * (math.sqrt(a - b))
            angle_change = math.degrees(dist_of_change / float(nradius))

            while abs(sum_of_angles) < (angle_bw_points - angle_change):
                new_angle = (angle_change * rot_val)
                sum_of_angles += new_angle
                new_bearing = centre_to_start + sum_of_angles
                delta_east, delta_north = calc_new_point(new_bearing, nradius)
                next_e = centre_point[0] + delta_east
                next_n = centre_point[1] + delta_north
                arc_point = [next_e, next_n]
                lines.append(arc_point)
            lines.append([end_e, end_n])

            # add heights
            if height_dif is not None:
                number_of_points = len(lines) - 1
                increment = height_dif / float(number_of_points)
                height = start_h
                new_lines = [(start_e, start_n, start_h)]
                for line in lines[1:-1]:
                    height += increment
                    line = (line[0], line[1], height)
                    new_lines.append(line)
                new_lines.append((end_e, end_n, end_h))

        else:

            lines = [start_neh,end_neh]

    except:
        lines = [start_neh, end_neh]


    return lines


# converts arc distance to chord
def arc2chord(distance, radi):
    p1 = distance / radi
    p2 = 2 * radi
    p3 = math.sin(p1 / 2)
    distance = p2 * p3
    return distance


# converts chord to arc
def chord2arc(distance, radi):
    if distance > (radi * 2):
        distance = radi * 2
    dsq = distance * distance
    inth = (radi * radi) - (dsq / 4)
    h = radi - math.sqrt(inth)
    fourh = 4 * h
    p1 = math.asin(distance / (h + (dsq / fourh)))
    p2 = (h + (dsq / fourh))
    distance = p1 * p2
    return distance


# calculates the centre point of a curve the angle between the points and the bearing to the centre of the arc
# returns the radius incase the modification of the coordinates has caused the radius to become smaller than
# half the chord
def calc_centre_point_of_curve(radius, bearing, distance, rot_val):
    half_chord = float(distance) / 2.0
    if half_chord > radius:
        radius = half_chord
    half_angle_bw_points = math.degrees(math.asin(float(half_chord) / float(radius)))
    angle_bw_points = half_angle_bw_points * 2
    angle_to_centre = 180 - (90 + half_angle_bw_points)
    bearing_to_centre = bearing + (angle_to_centre * rot_val)

    return angle_bw_points, bearing_to_centre, radius


def flip_rotation(rotdir):
    if rotdir.lower() == 'cw':
        rotdir = 'ccw'
    else:
        rotdir = 'cw'

    return rotdir


def calc_arc_length(start_ne, end_ne, centre_ne, rot, radius):
    start_ne = sg.Point(start_ne)
    end_ne = sg.Point(end_ne)
    centre_ne = sg.Point(centre_ne)
    chord = calc_distance(start_ne, end_ne)
    b2s = math.degrees(calc_bearing(start_ne, centre_ne,))
    b2e = math.degrees(calc_bearing(end_ne, centre_ne))
    angle = round(calc_inside_360(b2s - b2e))

    # lets not create the large arc here unless we recalc..
    if rot == 'cw':
        if angle >= 180 or angle == 0:
            large = False
        else:
            large = True
    else:
        if angle >= 180:
            large = True
        else:
            large = False
    circumference = math.pi * 2 * radius
    arc_length = chord2arc(chord, radius)

    if large is True:
        arc_length = circumference - arc_length
    return arc_length


# assigns a positive or negative value for a rotation direction
def calc_value_for_rotation(rot_dir):
    if rot_dir.lower() == 'ccw':
        rot_val = -1
    else:
        rot_val = 1
    return rot_val


def calc_area(coords, area=0):
    for prev, item, nxt in ts_functions.previous_and_next(coords):
        if nxt:
            (e1, e2) = item
            (n1, n2) = nxt
            a = (e1 * n2) - (e2 * n1)
            area += a
    return area


def get_poly_area(geom):
    area = 0
    if geom.type == 'Polygon':
        exterior_coords = geom.exterior.coords[:]
        area += calc_area(exterior_coords, 0)
        for interior in geom.interiors:
            interior_coords = interior.coords[:]
            area += calc_area(interior_coords, 0)
    elif geom.type == 'MultiPolygon':
        for part in geom:
            get_poly_area(part)
    else:
        area = 0
    return area


def dd2dms(x):
    degrees = float(int(x))  # float adds decimal point where x is whole
    tempminutes = (x-degrees)*60
    minutes = int(tempminutes)
    seconds = round((tempminutes-minutes)*60)  # round is needed here to determine if seconds are >=60
    if seconds >= 60:
        minutes = minutes + 1
        seconds = seconds - 60
    if minutes >= 60:
        degrees = degrees + 1
        minutes = minutes - 60
        if degrees > 360:
            degrees = degrees - 360

    minutes = float(minutes)/100.
    seconds = float(seconds)/10000.
    # removes float anomoly. Without round here, some float numbers
    # will revert back to base 2, example: 179.513 = 179.129999998
    dms = round(degrees + minutes + seconds,4)
    degree,minsec = str(dms).split('.')
    minsec = minsec.ljust(4,'0')
    dms = degree + '.' + minsec
    return dms


def dms2dd(x):
    # convert the float to a string for processing
    pos = True
    x = float(x)
    if x < 0:
        pos = False
    else:
        pos = True

    x = str(x) # float adds decimal point where missing from whole numbers
    degrees, minsec = x.split('.')
    minsec = minsec.ljust(4, '0')
    degrees = int(degrees)
    minutes = int(minsec[:2])

    power = 2-(len(minsec[2:]))
    seconds = (float(minsec[2:])) * (10**power)

    if pos is True:
        dd = degrees + float(minutes) / 60. + float(seconds)/3600.
    else:
        dd = degrees - float(minutes) / 60. - float(seconds) / 3600.

    return dd


# calcs plane distance between 2 points
def calc_distance(a, b):
    return a.distance(b)


# calcs plane bearing between 2 points
def calc_bearing(end, start):
    return math.atan2(end.x - start.x, end.y - start.y)


# calcs a plane coordinate deltas from a bearing and a distance
def calc_new_point(bearing, distance):
    delta_east = (math.sin(bearing / (180 / math.pi))) * distance
    delta_north = (math.cos(bearing / (180 / math.pi))) * distance

    return delta_east, delta_north


# adds or removes 360 if above or below 360 degress
def calc_inside_360(b):
    if b < 0:
        b += 360
    elif b >= 360:
        b -= 360
    return b
