"""
/***************************************************************************
 LandXML2QGIS
                                 A QGIS plugin
 Import Vic, NSW LandXML into QGIS temp layers
 Generated by Plugin Builder: http://g-sherman.github.io/Qgis-Plugin-Builder/
                              -------------------
        begin                : 2019-08-22
        git sha              : $Format:%H$
        copyright            : (C) 2019 by James Leversha, Department of Environment, Land, Water and Planning
        email                : james.k.leversha@delwp.vic.gov.au
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""

from collections import defaultdict
from .ts_calculations import calc_arc_length, calc_arc_values, dms2dd, arc2chord, calc_inside_360, \
     calc_bearing, calc_distance, calc_new_point
from .ts_functions import previous_and_next, split_point_list_into_chunks
import shapely.geometry as sg
try:
    from lxml import etree
    old_etree = False
except ImportError:
    old_etree = True
    try:
        import xml.etree.cElementTree as etree
    except ImportError:
        import xml.etree.ElementTree as etree

import math


def recalc_point(points, lines, arcs, is2point):
    import networkx as nx
    nodes = set()
    G = nx.Graph()
    for data in [lines, arcs]:
        for key, value in data.items():

            setup = is2point.get((value.get('@setupID').strip()))
            target = is2point.get((value.get('@targetSetupID').strip()))
            distance = value.get('@horizDistance')
            if distance is None:
                distance = value.get('@length')
                try:
                    distance = float(distance)
                    rad = float(value.get('@radius'))
                    distance = arc2chord(distance, rad)
                except ValueError or TypeError as err:
                    print(err)

            azimuth = value.get('@azimuth')
            if azimuth is None:
                azimuth = value.get('@chordAzimuth')
            try:
                distance = float(distance)
                azimuth = float(dms2dd(azimuth))
                if setup is not None and target is not None:
                    G.add_edge(setup, target, distance=distance, bearing=azimuth, st=(setup, target))
                    nodes.add(setup)
                    nodes.add(target)

            except ValueError or TypeError as err:
                print(err)

    # find unconnected points
    unconnected = []
    for point in points:
        if point not in nodes:
            unconnected.append(point)

    connected = nx.is_k_edge_connected(G, 1)
    if connected is False:
        branches = list(nx.k_edge_components(G, 1))
        max_b = max(branches, key=len)
        ref_point = sorted([item for item in max_b])[-1]
        for item in branches:
            if item != max_b:
                for node in item:
                    unconnected.append(node)
    else:

        ref_point = sorted(list(nodes))[-1]

    for point in unconnected:
        target = points.get(point)
        setup = points.get(ref_point)
        if setup is not None and target is not None:
            distance = calc_distance(setup, target)
            azimuth = calc_bearing(target, setup)
            azimuth = calc_inside_360(math.degrees(azimuth))
            G.add_edge(ref_point, point, distance=distance, bearing=azimuth, st=(ref_point, point))

    new_geom = {ref_point: points.get(ref_point)}
    short_paths = nx.single_source_shortest_path(G, ref_point)
    ref_point_coords = list(points.get(ref_point).coords)
    for key, value in short_paths.items():
        ref_e, ref_n = ref_point_coords[0]
        for prev, item, nxt in previous_and_next(value):
            if nxt is not None:
                b_d = G.get_edge_data(item, nxt)
                bearing = b_d['bearing']
                if (item, nxt) != b_d['st']:
                    bearing += 180
                    bearing = calc_inside_360(bearing)

                delta_e, delta_n = calc_new_point(bearing, b_d['distance'])
                ref_e += delta_e
                ref_n += delta_n

                if nxt not in new_geom:
                    new_geom[nxt] = sg.Point(ref_e, ref_n)

    return new_geom


# parses the xml and converts it to a python dictionary, haven't checked this fully
# maintains order for parcels as clockwise is important.
def etree_to_dict(t):
    # removes namespace from tag
    if not isinstance(t.tag, (tuple,)):
        if '}' in t.tag:
            t.tag = t.tag.split('}')[1]

    d = {t.tag: {} if t.attrib else None}
    children = list(t)
    if children:
        dd = defaultdict(list)
        # maintains the order on CoordGeom by adding a tuple number
        if t.tag == 'CoordGeom':
            for child in children:
                child.tag = child.tag.split('}')[1]

                child.text = str(children.index(child) + 1)

        for dc in map(etree_to_dict, children):
            for k, v in dc.items():
                dd[k].append(v)
        d = {t.tag: {k: v[0] if len(v) == 1 else v for k, v in dd.items()}}
    if t.attrib:
        d[t.tag].update(('@' + k, v)
                        for k, v in t.attrib.items())
    if t.text:
        text = t.text.strip()
        if children or t.attrib:
            if text:
                d[t.tag]['#text'] = text
        else:
            d[t.tag] = text

    if t.tag == 'CoordGeom':
        new_dict = {}
        items = list(d[t.tag].items())

        for item in items:
            if '@name' in item:
                new_dict[item[0]] = item[1]
                item_index = items.index(item)
                items.pop(item_index)

        # f = sorted(items, key=lambda x: x[0])
        new_list = []
        for item in items:
            if not isinstance(item[1], list):
                item = (item[0], [item[1]])
            for i in item[1]:
                new_list.append(((int(i['#text']), item[0]), i))

        f = sorted(new_list, key=lambda x: x[0])

        for item in f:
            line_order = new_dict.get('lineOrder', [])
            line_order.append({item[0][1]: item[1]})
            new_dict['lineOrder'] = line_order

        d[t.tag] = new_dict

    # change some lists to dicts using name/pntRef as reference for quick lookup
    # probably a nicer way to do this
    if t.tag in ['CgPoints', 'Parcels', 'ObservationGroup', 'Monuments']:
        for key, value in d[t.tag].items():
            if key in ['CgPoint', 'ReducedObservation', 'ReducedArcObservation', 'Parcel']:
                d[t.tag][key] = convert_to_name_dict(value)
            elif key in ['Monument']:
                d[t.tag][key] = convert_to_name_dict(value, attribute='@pntRef')
            elif key in ['RedHorizontalPosition']:
                d[t.tag][key] = convert_to_name_dict(value, attribute='@setupID')

    return d


def convert_to_name_dict(value, attribute='@name'):
    count = 0
    new_value = {}
    if not isinstance(value, (list,)):
        value = [value]
    for item in value:
        name = item.get(attribute)
        if name is None:
            count += 1
            name = 'no_name_' + str(count)
        count_2 = 0
        while name in new_value.keys():
            count_2 += 1
            if '_' in name:
                name, end = name.split('_')
            name += '_' + str(count_2) + 'DUP'
        new_value[name] = item
    return new_value


# get the point dictionary from xml2dict modified dictionary
# pull the text attribute and split it into enh list
# convert to floats
# load it into another dictionary with point name as reference
def point_geometry_dictionary(point_dictionary):
    point_geometry = {}
    for key, value in point_dictionary.items():
        enh = value.get('#text').split()
        if len(enh) == 3:
            enh = [enh[1]] + [enh[0]] # + [enh[2]]
        else:
            enh = [enh[1]] + [enh[0]]
        if enh is not None:
            point_geometry[key] = sg.Point((float(x) for x in enh))
            # """if len(enh) == 2:
            #     point_geometry[key]['east'] = enh[1]
            #     point_geometry[key]['north'] = enh[0]
            # elif len(enh) == 3:
            #     point_geometry[key]['east'] = enh[1]
            #     point_geometry[key]['north'] = enh[0]
            #     point_geometry[key]['height'] = enh[2]"""
    return point_geometry


# calculate the line and arc geometry from points, point references
# for each value in the line/arc dictionary, get the setup and target IS references
# convert these to the point references using the point reference dictionary
# lookup the geometry from the point geometry dictionary already created.
# if its an arc, calculate a psuedo arc
# store to new dictionary using same key as in the dictionary created from the xml
def line_geometry_dictionary(line_dictionary, point_geometry, ref_dict, arc=False, main_line='purpose', use_nx=True):
    point_references = ref_dict.get('is2point')
    line_geometry = {}
    arc_references = ref_dict.get('arcReferences', {})
    line_lookup = ref_dict.get('lineLookups', {})
    arc_lookup = ref_dict.get('arcLookups', {})
    if use_nx is True:
        import networkx as nx
        graph = ref_dict.get('graph', nx.Graph())
    else:
        graph = None
    for key, value in line_dictionary.items():
        setup = point_references.get(value.get('@setupID'))
        target = point_references.get(value.get('@targetSetupID'))
        setup_enh = point_geometry.get(setup, sg.Point([-999, -999]))
        target_enh = point_geometry.get(target, sg.Point([-999, -999]))
        main_line = '@' + main_line
        line_type = value.get(main_line)
        name = value.get('@name')

        if arc is False:
            chord_val = value.get('@horizDistance', 0)
            try:
                chord = float(chord_val)
            except ValueError:
                chord = 0
            dms_bearing = value.get('@azimuth', '0')
            try:
                float(dms_bearing)
            except ValueError:
                dms_bearing = '0'

            dd_bearing = dms2dd(dms_bearing)

            line_geometry[key] = sg.LineString([setup_enh, target_enh])

        else:
            rot = value.get('@rot')
            radius = value.get('@radius')
            arc_length = value.get('@length')
            try:
                chord = arc2chord(float(arc_length), float(radius))
            except:
                chord = 0
            dms_bearing = value.get('@chordAzimuth')
            dd_bearing = dms2dd(dms_bearing)
            # try to get an existing arc if it exists
            line = arc_references.get((setup, target))
            if line is None:
                rev_line = arc_references.get((target, setup))
                if rev_line is not None:
                    line = [x for x in reversed(rev_line.coords)]
                else:

                    line = calc_arc_values(setup_enh, target_enh, radius, rot, arc_length)
                    arc_references[(setup, target)] = sg.LineString(line)
            line_geometry[key] = sg.LineString(line)

            arc_lookup[(setup, target)] = {'@rot': rot, '@radius': float(radius), '@name': name}

        line_lookup[(setup, target)] = {'bearing': dd_bearing, 'distance': chord, main_line: line_type,
                                        '@name': name, 'reversed': False, 'arc': arc}
        rev_dd_bearing = calc_inside_360(dd_bearing + 180)
        line_lookup[(target, setup)] = {'bearing': rev_dd_bearing, 'distance': chord,
                                        main_line: line_type, '@name': name, 'reversed': True}
        if use_nx is True:
            graph.add_edge(setup, target, weight=chord)
    ref_dict['lineLookups'] = line_lookup
    ref_dict['arcReferences'] = arc_references
    ref_dict['arcLookups'] = arc_lookup
    ref_dict['graph'] = graph
    return line_geometry, ref_dict


# creates geometries for all parcels except for multipart
# multipart is left off because its parts have separate geometry anyway
# these parts can be called by accessing the multipart polygon from the main json dictionary
# and then calling their respective geometries and merging from this dictionary
# arc reference data is used in the creation of polygons so that no slivers or overlaps with lines
# and other parcels, if no arc reference data exists it will build its own arc
# however having this situation fails LRS schema checks
# irregularLines require a pointListReference in the reference dictionary to be created
# this pointListReference needs to be created so all irregular line points are in the point dictionary
# if an adjustment is then applied to the points, this means all of the polygon points will be adjusted
# simply by calling a point number rather than chasing coordinates.
# it also relies on line order being built in the original json, when converting to json
# it groups lines curves and irregular lines as 3 separate dictionaries, which in turn would cause it to
# lose its line order, some rough code has been created to avoid this issue
# this code probably can be cleaned up
def polygon_geometry_dictionary(polygon_data, point_geom_dict, reference_dict):
    point_reference_list = reference_dict.get('pointListReference', {})
    arc_reference_list = reference_dict.get('arcReferences', {})
    polygon_geometries = {}
    for key, value in polygon_data.items():
        if value.get('@parcelType') != 'Multipart':
            coord_geom = value.get('CoordGeom', {})
            line_order = coord_geom.get('lineOrder', [])
            polygon_name = key
            poly_geom = []
            s_e = []
            for item in line_order:
                for line_type, line_values in item.items():
                    start_point = line_values.get('Start', {}).get('@pntRef')
                    start_coord = point_geom_dict.get(start_point)

                    end_point = line_values.get('End', {}).get('@pntRef')
                    end_coord = point_geom_dict.get(end_point)
                    s_e.append([start_point, end_point])

                    if end_coord is not None and start_coord is not None:
                        start = list(start_coord.coords)[0]
                        end = list(end_coord.coords)[0]
                        if item == line_order[0]:
                            poly_geom.append(start_coord)
                        if line_type == 'Curve':
                            arc_values = arc_reference_list.get((start_point, end_point))
                            if arc_values is None:
                                arc_values = arc_reference_list.get((end_point, start_point))
                                if arc_values is not None:
                                    arc_values = sg.LineString([x for x in reversed(arc_values.coords[:])])
                            if arc_values is None:
                                centre_point = line_values.get('Center', {}).get('@pntRef')
                                centre_coord = point_geom_dict.get(centre_point)
                                centre = list(centre_coord.coords)[0]
                                rot = line_values.get('@rot')
                                radius = float(line_values.get('@radius'))
                                arc_length = calc_arc_length(start, end, centre, rot, radius)

                                arc_values = sg.LineString(
                                    calc_arc_values(start_coord, end_coord, radius, rot, arc_length))
                                arc_reference_list[(start_point, end_point)] = arc_values
                            for item in arc_values.coords[1:]:
                                poly_geom.append(sg.Point(item))
                        elif line_type == 'IrregularLine':
                            pnt_list_2D = line_values.get('PntList2D', [])
                            pnt_list_3D = line_values.get('PntList3D', [])
                            if pnt_list_2D is not None:
                                point_order = point_reference_list.get(pnt_list_2D, [])
                                if len(point_order) > 0:
                                    if point_order[0] == (start_point[0], start_point[1]):
                                        point_order = point_order[1:]
                                    if point_order[-1] == (end_point[0], end_point[1]):
                                        point_order = point_order[:-1]

                            elif pnt_list_3D is not None:
                                point_order = point_reference_list.get(pnt_list_3D, [])
                                if len(point_order) > 0:
                                    if point_order[0] == start_point:
                                        point_order = point_order[1:]
                                    if point_order[-1] == end_point:
                                        point_order = point_order[:-1]
                            else:
                                point_order = []
                            for point in point_order:
                                coord = point_geom_dict.get(point)
                                poly_geom.append(coord)
                        else:
                            poly_geom.append(end_coord)

            # force the polygon shut
            if len(poly_geom) > 1:
                if poly_geom[0] != poly_geom[-1]:
                    poly_geom.append(poly_geom[0])

            # fake the geometry here, who knows why victoria decided to have polygons with only one line...
            elif len(poly_geom) == 1:
                poly_geom.append(poly_geom[0])
                poly_geom.append(poly_geom[0])

            poly_geom = [[p.x, p.y] for p in poly_geom]

            polygon_geom = sg.Polygon(poly_geom)
            if polygon_geom.is_valid is False and (s_e[0] == s_e[-1]) and len(s_e) > 1:
                polygon_geom = polygon_geom.buffer(0)
            polygon_geometries[polygon_name] = polygon_geom

    return point_geom_dict, polygon_geometries


# creates a representation of the geometry for points, lines, polygons, and arcs.
def create_geometry_dictionary(xml_in_dict, reference_dict, recalc=True, main_line='VIC'):
    is2point = reference_dict.get('is2point')
    point_data = xml_in_dict.get('LandXML', {}).get('CgPoints', {}).get('CgPoint', {})
    line_data = xml_in_dict.get('LandXML', {}).get('Survey', {}).get('ObservationGroup', {}).get('ReducedObservation',
                                                                                                 {})
    arc_data = xml_in_dict.get('LandXML', {}).get('Survey', {}).get('ObservationGroup', {}).get('ReducedArcObservation',
                                                                                                {})
    polygon_data = xml_in_dict.get('LandXML', {}).get('Parcels', {}).get('Parcel', {})
    geometry_dictionary = {}
    # points
    points = point_geometry_dictionary(point_data)
    geometry_dictionary['points'] = points

    # add irregular lines to geometry
    point_list_reference = reference_dict.get('pointListReference', {})
    points, point_list_reference = point_list_to_pnt_reference(points, polygon_data, point_list_reference)
    reference_dict['pointListReference'] = point_list_reference

    # recalc geometry
    if recalc is True:
        points = recalc_point(points, line_data, arc_data, is2point)
    geometry_dictionary['points'] = points
    # lines
    geometry_dictionary['lines'], reference_dict = line_geometry_dictionary(line_data, points,
                                                                            reference_dict, use_nx=recalc,
                                                                            main_line=main_line)
    # arcs
    geometry_dictionary['arcs'], reference_dict = line_geometry_dictionary(arc_data, points, reference_dict, arc=True,
                                                                           use_nx=recalc, main_line=main_line)

    # polygons
    # this will add information to the point geom dictionary if required for irregular lines

    geometry_dictionary['points'], geometry_dictionary['polygons'] = polygon_geometry_dictionary(polygon_data, points,
                                                                                                 reference_dict)

    return geometry_dictionary, reference_dict


# converts a point list of 2D or 3D coordinates to actual point references,
# this needs to be done so that if an adjustment is made to the entire file these then can be
# applied to polygons that refernce these marks rather than trying to apply it on the fly.
# this needs a clean up
def coordinate_2D_3D_reference(point_geom_dict):
    # create 2 dictionaries of coordinates: point pairs that already exist for both 2D and 3D
    current_coords_2D = {}
    current_coords_3D = {}
    for key, value in point_geom_dict.items():
        coords = list(value.coords)[0]
        if len(coords) == 2:
            e, n = coords
            h = 0
            current_coords_2D[(e, n)] = key
            current_coords_3D[(e, n, h)] = key
        elif len(coords) == 3:
            n, e, h = coords
            current_coords_2D[(e, n)] = key
            current_coords_3D[(e, n, h)] = key

    return current_coords_2D, current_coords_3D


def get_all_point_lists(polygon_data):
    pnt_lists_2D = []
    pnt_lists_3D = []
    for key, value in polygon_data.items():
        coord_geom = value.get('CoordGeom', {})
        line_order = coord_geom.get('lineOrder', [])
        for item in line_order:
            for line_type, line_values in item.items():
                if line_type == 'IrregularLine':
                    pnt_list = line_values.get('PntList2D')
                    if pnt_list is not None:
                        pnt_lists_2D.append(pnt_list)
                    else:
                        pnt_list = line_values.get('PntList3D')
                        if pnt_list is not None:
                            pnt_lists_3D.append(pnt_list)
    return pnt_lists_2D, pnt_lists_3D


def point_list_to_pnt_reference(point_geom_dict, polygon_data, pnt_list_references=None):
    if pnt_list_references is None:
        pnt_list_references = {}
    current_coords_2D, current_coords_3D = coordinate_2D_3D_reference(point_geom_dict)
    pnt_lists_2D, pnt_lists_3D = get_all_point_lists(polygon_data)
    pnt_dicts = [(2, current_coords_2D, pnt_lists_2D), (3, current_coords_3D, pnt_lists_3D)]
    count = 0

    for value in pnt_dicts:

        size = value[0]
        current_coords = value[1]
        pnt_lists = value[2]
        for key in pnt_lists:
            pnt_list = split_point_list_into_chunks(key, size)
            value_list = []
            for item in pnt_list:

                if size == 2:
                    item = (item[1], item[0])
                elif size == 3:
                    item = (item[1], item[0], item[2])

                value = current_coords.get(item)
                if value is None:
                    count += 1
                    value = 'PntList-' + str(count)
                    while value in point_geom_dict.keys():
                        count += 1
                        value = 'PntList-' + str(count)
                    point_geom_dict[value] = sg.Point(item)
                    current_coords[item] = value
                    value_list.append(value)
                else:
                    value_list.append(value)

            if len(value_list) > 0:

                pnt_list_references[key] = value_list

    return point_geom_dict, pnt_list_references


# generates a control dictionary from all the areas in a file that it is located
def generate_control(xml_dictionary, reference_dictionary):
    pass


# checks: to see if document matches schema
# inputs: xml, schema
# outputs a full list of dictionaries of schema failures.
def schema_failures(root, schema):
    def result(item, error, message, value, item_type, error_code):
        return dict(item=item, qa_error=error, message=message,
                    value=value, item_type=item_type, error_code=error_code)

    xml_schema_doc = etree.parse(schema)
    xmlschema = etree.XMLSchema(xml_schema_doc)
    xmlschema.validate(root)
    log = xmlschema.error_log

    errors = []
    for item in log:
        error = 'P1'
        message = result('Value at Line: ' + str(item.line), error, item.message, item.type_name, 'Schema', 'TS99')
        errors.append(message)

    return errors


# this function creates a dictionary representation of a land xml
def create_dictionaries_from_xml(xml, recalc=True, region='VIC'):
    main_line = ''
    if region == 'VIC':
        main_line = 'purpose'
    elif region == 'NSW':
        main_line = 'desc'
    elif region == 'QLD':
        main_line = 'some_value'
    if old_etree is True:
        tree = etree.parse(xml)
        root = tree.getroot()

    else:
        parser = etree.XMLParser(remove_comments=True)
        try:
            if not isinstance(xml, str):
                xml = xml.get()['Body']
            root = etree.fromstring(xml, parser)
            etree.strip_tags(root, etree.Comment)

        except Exception as err:
            tree = etree.parse(xml, parser=parser)
            root = tree.getroot()

    errs = {}
    xml_in_a_dict = etree_to_dict(root)

    # lookup dictionaries
    ins = xml_in_a_dict.get('LandXML', {}).get('Survey', {}).get('InstrumentSetup', {})
    is_to_point_reference_dict = {item.get('@id'): item.get('InstrumentPoint', {}).get('@pntRef') for item in ins}
    point_to_is_reference_dict = {value: key for key, value in is_to_point_reference_dict.items()}
    reference_dicts = {'is2point': is_to_point_reference_dict,
                       'point2is': point_to_is_reference_dict,
                       'schemaChecks': errs}
    geometry_dict, reference_dicts = create_geometry_dictionary(xml_in_a_dict, reference_dicts,
                                                                recalc=recalc, main_line=main_line)

    return xml_in_a_dict, reference_dicts, geometry_dict
