# -*- coding: utf-8 -*-
"""
/***************************************************************************
 LandXML2QGIS
                                 A QGIS plugin
 Import Vic, NSW LandXML into QGIS temp layers
 Generated by Plugin Builder: http://g-sherman.github.io/Qgis-Plugin-Builder/
                              -------------------
        begin                : 2019-08-22
        git sha              : $Format:%H$
        copyright            : (C) 2019 by James Leversha, Department of Environment, Land, Water and Planning
        email                : james.k.leversha@delwp.vic.gov.au
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""
from qgis.PyQt.QtCore import QSettings, QTranslator, QCoreApplication
from qgis.PyQt.QtGui import QIcon
from qgis.PyQt.QtWidgets import QAction
from qgis.core import QgsGeometry, QgsField, QgsVectorLayer, QgsFeature, QgsProject, QgsCoordinateReferenceSystem
from qgis.PyQt.QtGui import QIcon
from qgis.PyQt.QtWidgets import QAction, QFileDialog
from qgis.PyQt import QtWidgets
from qgis.PyQt import QtCore
from PyQt5.QtCore import QVariant

import sys

sys.path.append("..")
from .dcmvr.ts_functions import deep_get
from .dcmvr.ts_readers import create_dictionaries_from_xml as cdfx

# Initialize Qt resources from file resources.py
from .resources import *
# Import the code for the dialog
from .landxml2qgis_dialog import LandXML2QGISDialog
import os.path


class LandXML2QGIS:
    """QGIS Plugin Implementation."""

    def __init__(self, iface):
        """Constructor.

        :param iface: An interface instance that will be passed to this class
            which provides the hook by which you can manipulate the QGIS
            application at run time.
        :type iface: QgsInterface
        """
        # Save reference to the QGIS interface
        self.iface = iface
        # initialize plugin directory
        self.plugin_dir = os.path.dirname(__file__)
        # initialize locale
        locale = QSettings().value('locale/userLocale')[0:2]
        locale_path = os.path.join(
            self.plugin_dir,
            'i18n',
            'LandXML2QGIS_{}.qm'.format(locale))

        if os.path.exists(locale_path):
            self.translator = QTranslator()
            self.translator.load(locale_path)
            QCoreApplication.installTranslator(self.translator)

        # Declare instance attributes
        self.actions = []
        self.menu = self.tr(u'&LandXML2QGIS')

        # Check if plugin was started the first time in current QGIS session
        # Must be set in initGui() to survive plugin reloads
        self.first_start = None

    # noinspection PyMethodMayBeStatic
    def tr(self, message):
        """Get the translation for a string using Qt translation API.

        We implement this ourselves since we do not inherit QObject.

        :param message: String for translation.
        :type message: str, QString

        :returns: Translated version of message.
        :rtype: QString
        """
        # noinspection PyTypeChecker,PyArgumentList,PyCallByClass
        return QCoreApplication.translate('LandXML2QGIS', message)


    def add_action(
        self,
        icon_path,
        text,
        callback,
        enabled_flag=True,
        add_to_menu=True,
        add_to_toolbar=True,
        status_tip=None,
        whats_this=None,
        parent=None):
        """Add a toolbar icon to the toolbar.

        :param icon_path: Path to the icon for this action. Can be a resource
            path (e.g. ':/plugins/foo/bar.png') or a normal file system path.
        :type icon_path: str

        :param text: Text that should be shown in menu items for this action.
        :type text: str

        :param callback: Function to be called when the action is triggered.
        :type callback: function

        :param enabled_flag: A flag indicating if the action should be enabled
            by default. Defaults to True.
        :type enabled_flag: bool

        :param add_to_menu: Flag indicating whether the action should also
            be added to the menu. Defaults to True.
        :type add_to_menu: bool

        :param add_to_toolbar: Flag indicating whether the action should also
            be added to the toolbar. Defaults to True.
        :type add_to_toolbar: bool

        :param status_tip: Optional text to show in a popup when mouse pointer
            hovers over the action.
        :type status_tip: str

        :param parent: Parent widget for the new action. Defaults None.
        :type parent: QWidget

        :param whats_this: Optional text to show in the status bar when the
            mouse pointer hovers over the action.

        :returns: The action that was created. Note that the action is also
            added to self.actions list.
        :rtype: QAction
        """

        icon = QIcon(icon_path)
        action = QAction(icon, text, parent)
        action.triggered.connect(callback)
        action.setEnabled(enabled_flag)

        if status_tip is not None:
            action.setStatusTip(status_tip)

        if whats_this is not None:
            action.setWhatsThis(whats_this)

        if add_to_toolbar:
            # Adds plugin icon to Plugins toolbar
            self.iface.addToolBarIcon(action)

        if add_to_menu:
            self.iface.addPluginToVectorMenu(
                self.menu,
                action)

        self.actions.append(action)

        return action

    def initGui(self):
        """Create the menu entries and toolbar icons inside the QGIS GUI."""

        icon_path = ':/plugins/landxml2qgis/icon.png'
        self.add_action(
            icon_path,
            text=self.tr(u'LandXML2QGIS'),
            callback=self.run,
            parent=self.iface.mainWindow())

        # will be set False in run()
        self.first_start = True


    def unload(self):
        """Removes the plugin menu item and icon from QGIS GUI."""
        for action in self.actions:
            self.iface.removePluginVectorMenu(
                self.tr(u'&LandXML2QGIS'),
                action)
            self.iface.removeToolBarIcon(action)

    def get_zone(self):
        zone_dict = {'MGA202054': '7854', 'MGA202055': '7855', 'MGA202056': '7856',
                     'MGA9454': '28354', 'MGA9455': '28355', 'MGA9456': '28356',
                     'AGD6654': '20254', 'AGD6655': '20255', 'AGD6656': '20256',
                     'ISG54-2': '102072', 'ISG54-3': '102073', 'ISG55-1': '102074',
                     'ISG55-2': '102075', 'ISG55-3': '102076', 'ISG56-1': '102077',
                     'ISG56-2': '102078', 'ISG56-3': '102079', 'Local': '7855'
                     }
        zone = self.dlg.zoneComboBox.currentText()
        datum = self.dlg.datumComboBox.currentText()
        return zone_dict.get(datum + zone, '7855')

    def create_temp_layer(self, type_data, layer_name, data, geometry, diagram, location='memory'):
        dz = self.get_zone()
        vl = QgsVectorLayer(type_data + '?crs=EPSG:' + dz, diagram + '_' + layer_name, location)
        pr = vl.dataProvider()

        for key, value in geometry.items():
            data_v = data.get(key, {})
            data_v['geometry'] = QgsGeometry.fromWkt(value.wkt)
            data_v['diagram'] = diagram
            data[key] = data_v

        columns = []
        for value in data.values():
            keys = [k for k in value if k not in columns]
            columns += keys
        columns = ['key'] + sorted(columns)
        items = [QgsField(item, QVariant.String) for item in columns]
        pr.addAttributes(items)

        vl.updateFields()

        for key, value in data.items():
            f = QgsFeature()
            f.setGeometry(value.get('geometry'))
            row = [key] + [value.get(k) for k in columns if k != 'key']
            f.setAttributes(row)
            pr.addFeature(f)

        vl.updateExtents()

        QgsProject.instance().addMapLayer(vl)

    def create_point_layer(self, xid, geom, xml_name):
        data = deep_get(xid, 'LandXML.CgPoints.CgPoint')
        geom = geom.get('points')
        self.create_temp_layer('Point', 'points', data, geom, xml_name)

    def create_line_layer(self, xid, geom, xml_name):
        data = deep_get(xid, 'LandXML.Survey.ObservationGroup.ReducedObservation')
        geom = geom.get('lines')
        self.create_temp_layer('LineString', 'lines', data, geom, xml_name)

    def create_arc_layer(self, xid, geom, xml_name):
        data = deep_get(xid, 'LandXML.Survey.ObservationGroup.ReducedArcObservation')
        geom = geom.get('arcs')
        self.create_temp_layer('LineString', 'arcs', data, geom, xml_name)

    def create_polygon_layer(self, xid, geom, xml_name):
        data = deep_get(xid, 'LandXML.Parcels.Parcel')
        geom = geom.get('polygons')
        self.create_temp_layer('Polygon', 'polygons', data, geom, xml_name)

    def run(self):
        """Run method that performs all the real work"""

        # Create the dialog with elements (after translation) and keep reference
        # Only create GUI ONCE in callback, so that it will only load when the plugin is started
        if self.first_start == True:
            self.first_start = False
            self.dlg = LandXML2QGISDialog()

        # show the dialog
        self.dlg.show()
        # Run the dialog event loop
        result = self.dlg.exec_()
        # See if OK was pressed
        if result:
            fname = self.dlg.lineEdit.text()
            region = str(self.dlg.regionComboBox.currentText())
            if self.dlg.recalcCheckBox.isChecked():
                recalc = True
            else:
                recalc = False
            xid, ref, geom = cdfx(fname, recalc=recalc, region=region)
            if self.dlg.pointCheckBox.isChecked() is True:
                if len(geom['points']) > 0:
                    self.create_point_layer(xid, geom, os.path.split(fname)[-1][:-4])
            if self.dlg.arcCheckBox.isChecked() is True:
                if len(geom['arcs']) > 0:
                    self.create_arc_layer(xid, geom, os.path.split(fname)[-1][:-4])
            if self.dlg.lineCheckBox.isChecked() is True:
                if len(geom['lines']) > 0:
                    self.create_line_layer(xid, geom, os.path.split(fname)[-1][:-4])
            if self.dlg.polygonCheckBox.isChecked() is True:
                if len(geom['polygons']) > 0:
                    self.create_polygon_layer(xid, geom, os.path.split(fname)[-1][:-4])
            # Do something useful here - delete the line containing pass and
            # substitute with your code.
